@@ -1,7 +1,7 @@
-/** @file weight.h
+/** @file
  * @brief Weighting scheme API.
  */
-/* Copyright (C) 2004,2007,2008,2009,2010,2011,2012,2015,2016 Olly Betts
+/* Copyright (C) 2004,2007,2008,2009,2010,2011,2012,2015,2016,2019 Olly Betts
  * Copyright (C) 2009 Lemur Consulting Ltd
  * Copyright (C) 2013,2014 Aarsh Shah
  * Copyright (C) 2016 Vivek Pal
@@ -63,7 +63,12 @@ class XAPIAN_VISIBILITY_DEFAULT Weight {
 	/// Sum of wdf over the whole collection for the current term.
 	COLLECTION_FREQ = 4096,
 	/// Number of unique terms in the current document.
-	UNIQUE_TERMS = 8192
+	UNIQUE_TERMS = 8192,
+	/** Sum of lengths of all documents in the collection.
+	 *
+	 *  This gives the total number of term occurrences.
+	 */
+	TOTAL_LENGTH = COLLECTION_SIZE | AVERAGE_LENGTH
     } stat_flags;
 
     /** Tell Xapian that your subclass will want a particular statistic.
@@ -254,6 +259,11 @@ class XAPIAN_VISIBILITY_DEFAULT Weight {
     /** @private @internal Initialise this object to calculate weights for term
      *  @a term.
      *
+     *  Old version of method, as used by 1.4.18 and earlier.  This
+     *  should only be referenced from inside the library and 1.4.19 and
+     *  later will call the new version instead.  We continue to provide it
+     *  mainly to avoid triggering ABI checking tools.
+     *
      *  @param stats	  Source of statistics.
      *  @param query_len_ Query length.
      *  @param term	  The term for the new object.
@@ -264,6 +274,23 @@ class XAPIAN_VISIBILITY_DEFAULT Weight {
 	       const std::string & term, Xapian::termcount wqf_,
 	       double factor);
 
+    /** @private @internal Initialise this object to calculate weights for term
+     *  @a term.
+     *
+     *  @param stats	  Source of statistics.
+     *  @param query_len_ Query length.
+     *  @param term	  The term for the new object.
+     *  @param wqf_	  The within-query-frequency of @a term.
+     *  @param factor	  Any scaling factor (e.g. from OP_SCALE_WEIGHT).
+     *  @param postlist   Pointer to a LeafPostList for the term (cast to void*
+     *			  to avoid needing to forward declare class
+     *			  LeafPostList in public API headers) which can be used
+     *			  to get wdf upper bound
+     */
+    void init_(const Internal & stats, Xapian::termcount query_len_,
+	       const std::string & term, Xapian::termcount wqf_,
+	       double factor, void* postlist);
+
     /** @private @internal Initialise this object to calculate weights for a
      *  synonym.
      *
@@ -315,6 +342,15 @@ class XAPIAN_VISIBILITY_DEFAULT Weight {
 	return stats_needed & UNIQUE_TERMS;
     }
 
+    /// @private @internal Test if this is a BoolWeight object.
+    bool is_bool_weight_() const {
+	// Checking the name isn't ideal, but (get_maxpart() == 0.0) isn't
+	// required to work without init() having been called.  We can at
+	// least avoid the virtual method call in most non-BoolWeight cases
+	// as most other classes will need at least some stats.
+	return stats_needed == 0 && name() == "Xapian::BoolWeight";
+    }
+
   protected:
     /** Don't allow copying.
      *
@@ -372,6 +408,11 @@ class XAPIAN_VISIBILITY_DEFAULT Weight {
     Xapian::termcount get_wdf_upper_bound() const {
 	return wdf_upper_bound_;
     }
+
+    /// Total length of all documents in the collection.
+    Xapian::totallength get_total_length() const {
+	return Xapian::totallength(average_length_ * collection_size_ + 0.5);
+    }
 };
 
 /** Class implementing a "boolean" weighting scheme.
@@ -451,7 +492,7 @@ class XAPIAN_VISIBILITY_DEFAULT TfIdfWeight : public Weight {
      *         which are indexed by the term t.
      *     @li 'p': Prob    idfn=log((N-Termfreq)/Termfreq)
      *     @li 'f': Freq    idfn=1/Termfreq
-     *     @li 's': Squared idfn=log(N/Termfreq)^2
+     *     @li 's': Squared idfn=(log(N/Termfreq))²
      *
      * @li The third and the final character indicates the normalization for
      *     the document weight.  The following normalizations are currently
@@ -812,7 +853,7 @@ class XAPIAN_VISIBILITY_DEFAULT InL2Weight : public Weight {
   public:
     /** Construct an InL2Weight.
      *
-     *  @param c  A non-negative and non zero parameter controlling the extent
+     *  @param c  A strictly positive parameter controlling the extent
      *		  of the normalization of the wdf to the document length. The
      *		  default value of 1 is suitable for longer queries but it may
      *		  need to be changed for shorter queries. For more information,
@@ -884,7 +925,7 @@ class XAPIAN_VISIBILITY_DEFAULT IfB2Weight : public Weight {
   public:
     /** Construct an IfB2Weight.
      *
-     *  @param c  A non-negative and non zero parameter controlling the extent
+     *  @param c  A strictly positive parameter controlling the extent
      *		  of the normalization of the wdf to the document length. The
      *		  default value of 1 is suitable for longer queries but it may
      *		  need to be changed for shorter queries. For more information,
@@ -957,7 +998,7 @@ class XAPIAN_VISIBILITY_DEFAULT IneB2Weight : public Weight {
   public:
     /** Construct an IneB2Weight.
      *
-     *  @param c  A non-negative and non zero parameter controlling the extent
+     *  @param c  A strictly positive parameter controlling the extent
      *		  of the normalization of the wdf to the document length. The
      *		  default value of 1 is suitable for longer queries but it may
      *		  need to be changed for shorter queries. For more information,
@@ -1031,7 +1072,7 @@ class XAPIAN_VISIBILITY_DEFAULT BB2Weight : public Weight {
   public:
     /** Construct a BB2Weight.
      *
-     *  @param c  A non-negative and non zero parameter controlling the extent
+     *  @param c  A strictly positive parameter controlling the extent
      *		  of the normalization of the wdf to the document length. A
      *		  default value of 1 is suitable for longer queries but it may
      *		  need to be changed for shorter queries. For more information,
@@ -1103,15 +1144,14 @@ class XAPIAN_VISIBILITY_DEFAULT DLHWeight : public Weight {
 
   public:
     DLHWeight() {
-	need_stat(AVERAGE_LENGTH);
 	need_stat(DOC_LENGTH);
-	need_stat(COLLECTION_SIZE);
 	need_stat(COLLECTION_FREQ);
 	need_stat(WDF);
 	need_stat(WQF);
 	need_stat(WDF_MAX);
 	need_stat(DOC_LENGTH_MIN);
 	need_stat(DOC_LENGTH_MAX);
+	need_stat(TOTAL_LENGTH);
     }
 
     std::string name() const;
@@ -1176,7 +1216,7 @@ class XAPIAN_VISIBILITY_DEFAULT PL2Weight : public Weight {
   public:
     /** Construct a PL2Weight.
      *
-     *  @param c  A non-negative and non zero parameter controlling the extent
+     *  @param c  A strictly positive parameter controlling the extent
      *		  of the normalization of the wdf to the document length. The
      *		  default value of 1 is suitable for longer queries but it may
      *		  need to be changed for shorter queries. For more information,
@@ -1246,7 +1286,7 @@ class XAPIAN_VISIBILITY_DEFAULT PL2PlusWeight : public Weight {
   public:
     /** Construct a PL2PlusWeight.
      *
-     *  @param c  A non-negative and non zero parameter controlling the extent
+     *  @param c  A strictly positive parameter controlling the extent
      *		  of the normalization of the wdf to the document length. The
      *		  default value of 1 is suitable for longer queries but it may
      *		  need to be changed for shorter queries. For more information,
@@ -1328,15 +1368,14 @@ class XAPIAN_VISIBILITY_DEFAULT DPHWeight : public Weight {
   public:
     /** Construct a DPHWeight. */
     DPHWeight() {
-	need_stat(AVERAGE_LENGTH);
 	need_stat(DOC_LENGTH);
-	need_stat(COLLECTION_SIZE);
 	need_stat(COLLECTION_FREQ);
 	need_stat(WDF);
 	need_stat(WQF);
 	need_stat(WDF_MAX);
 	need_stat(DOC_LENGTH_MIN);
 	need_stat(DOC_LENGTH_MAX);
+	need_stat(TOTAL_LENGTH);
     }
 
     std::string name() const;
@@ -1433,9 +1472,7 @@ class XAPIAN_VISIBILITY_DEFAULT LMWeight : public Weight {
 	    else
 		param_smoothing2 = 0.05;
 	}
-	need_stat(AVERAGE_LENGTH);
 	need_stat(DOC_LENGTH);
-	need_stat(COLLECTION_SIZE);
 	need_stat(RSET_SIZE);
 	need_stat(TERMFREQ);
 	need_stat(RELTERMFREQ);
@@ -1443,6 +1480,7 @@ class XAPIAN_VISIBILITY_DEFAULT LMWeight : public Weight {
 	need_stat(WDF);
 	need_stat(WDF_MAX);
 	need_stat(COLLECTION_FREQ);
+	need_stat(TOTAL_LENGTH);
 	if (select_smoothing == ABSOLUTE_DISCOUNT_SMOOTHING)
 	    need_stat(UNIQUE_TERMS);
 	if (select_smoothing == DIRICHLET_PLUS_SMOOTHING)
